const NEEDS_WRAP_RE = /^(\(.*\)|\\?.)$/;
const wrap = (s) => {
  const v = s.toString();
  return NEEDS_WRAP_RE.test(v) ? v : `(${v})`;
};

const createInput = (s) => {
  return {
    toString: () => s.toString(),
    and: Object.assign((input) => createInput(`${s}${exactly(input)}`), {
      referenceTo: (groupName) => createInput(`${s}\\k<${groupName}>`)
    }),
    or: (input) => createInput(`(${s}|${exactly(input)})`),
    after: (input) => createInput(`(?<=${exactly(input)})${s}`),
    before: (input) => createInput(`${s}(?=${exactly(input)})`),
    notAfter: (input) => createInput(`(?<!${exactly(input)})${s}`),
    notBefore: (input) => createInput(`${s}(?!${exactly(input)})`),
    times: Object.assign((number) => createInput(`${wrap(s)}{${number}}`), {
      any: () => createInput(`${wrap(s)}*`),
      atLeast: (min) => createInput(`${wrap(s)}{${min},}`),
      between: (min, max) => createInput(`${wrap(s)}{${min},${max}}`)
    }),
    optionally: () => createInput(`${wrap(s)}?`),
    as: (key) => createInput(`(?<${key}>${s})`),
    at: {
      lineStart: () => createInput(`^${s}`),
      lineEnd: () => createInput(`${s}$`)
    }
  };
};

const charIn = (chars) => createInput(`[${chars.replace(/[-\\^\]]/g, "\\$&")}]`);
const charNotIn = (chars) => createInput(`[^${chars.replace(/[-\\^\]]/g, "\\$&")}]`);
const anyOf = (...args) => createInput(`(${args.map((a) => exactly(a)).join("|")})`);
const char = createInput(".");
const word = createInput("\\b\\w+\\b");
const wordChar = createInput("\\w");
const wordBoundary = createInput("\\b");
const digit = createInput("\\d");
const whitespace = createInput("\\s");
const letter = createInput("[a-zA-Z]");
const tab = createInput("\\t");
const linefeed = createInput("\\n");
const carriageReturn = createInput("\\r");
const not = {
  wordChar: createInput("\\W"),
  wordBoundary: createInput("\\B"),
  digit: createInput("\\D"),
  whitespace: createInput("\\S"),
  letter: createInput("[^a-zA-Z]"),
  tab: createInput("[^\\t]"),
  linefeed: createInput("[^\\n]"),
  carriageReturn: createInput("[^\\r]")
};
const maybe = (str) => createInput(`${wrap(exactly(str))}?`);
const exactly = (input) => typeof input === "string" ? createInput(input.replace(/[.*+?^${}()|[\]\\/]/g, "\\$&")) : input;
const oneOrMore = (str) => createInput(`${wrap(exactly(str))}+`);

const withIndices = "d";
const caseInsensitive = "i";
const global = "g";
const multiline = "m";
const dotAll = "s";
const unicode = "u";
const sticky = "y";

const createRegExp = (raw, flags) => new RegExp(exactly(raw).toString(), [...flags || ""].join(""));

export { anyOf, carriageReturn, caseInsensitive, char, charIn, charNotIn, createRegExp, digit, dotAll, exactly, global, letter, linefeed, maybe, multiline, not, oneOrMore, sticky, tab, unicode, whitespace, withIndices, word, wordBoundary, wordChar };
